// ExecuteTime.cpp : Implementation file for MQSI V2.0.1 plug-in node
//
// Template built by MQSI V2 Visual Studio Wizard (kolban@us.ibm.com)
//
// Installation:
// For NT:		<MQSI Root>\bin
// For Unix:	<MQSI Root>/lil
//
#ifdef _WIN32
#include "stdafx.h"
#endif

#ifndef _WIN32
#include <sys/time.h>
#endif

#include <stdarg.h>		// Required for variable length arguments
#include <stdio.h>		// Required for file IO and NULL declaration
#include <stdlib.h>		// Required for memory manipulation
#include <BipCci.h>		// Common functions (between nodes and parsers)
#include <BipCni.h>		// Message processing nodes

#include "NodeUtils.h"	// Include the required utility functions

#ifdef _WIN32
#include "NT_msg/ExecuteTime_msg.h"
#else
#include "AIX_msg/EXECUTETIME_msg.h"
#endif


enum MeasureType {TOEND, TIMESTART, TIMEEND};

///////////////////////////////////////////////////////////
// NodeContext
struct NodeContext
{
	MQSIX_ATTRIBUTES	*attributes;
	MQSIX_NODE_INFO		brokerInfo;
	CciTerminal			*inTerminal;	// In Terminal
	CciTerminal			*outTerminal;	// Out Terminal
	enum MeasureType	iMeasureType;
	char				strEndLabelU[MQSIX_MAX_STRING];
	CciChar				*label;
};

///////////////////////////////////////////////////////////
// MY_TIME_T
// The definition of this data type is platform specific.
//
#ifdef _WIN32
typedef unsigned __int64 MY_TIME_T;
#else
typedef struct timeval MY_TIME_T;
#endif

#define MAX_TIME_ENTRIES (50)

typedef struct
{
	char		name[MQSIX_MAX_STRING];
	MY_TIME_T	time;
	int			used;
} NODE_TIME_ENTRY;

NODE_TIME_ENTRY	entryList[MAX_TIME_ENTRIES];

#ifdef _WIN32
HMODULE	g_module;
BOOL APIENTRY DllMain(
	HANDLE hModule,
	DWORD  ul_reason_for_call,
	LPVOID lpReserved
					 )
{
	g_module = hModule;
    return TRUE;
}

void __declspec(dllexport) *bipGetMessageflowNodeFactory();

#endif

int OS_GetPid()
{
#ifdef _WIN32
	return(GetCurrentProcessId());
#else
	return(getpid());
#endif
}

int OS_GetThread()
{
#ifdef _WIN32
	return(GetCurrentThreadId());
#else
	return(pthread_self());
#endif
}
///////////////////////////////////////////////////////////
// CreateNodeContext
static void *CreateNodeContext(
	CciFactory	*factoryObject,
	CciChar		*nodeName,
	CciNode		*nodeObject)
{
	struct NodeContext	*pMyContext;
	int					iRc;

	pMyContext = (struct NodeContext *)malloc(sizeof(struct NodeContext));

	mqsixLog(">> CreateNodeContext: Node Name=%S\n", nodeName);
	pMyContext->inTerminal = cniCreateInputTerminal(&iRc, nodeObject, L"in");
	pMyContext->outTerminal = cniCreateOutputTerminal(&iRc, nodeObject, L"out");
	pMyContext->attributes = NULL;
	pMyContext->label = NULL;
/*
	LARGE_INTEGER li;
	if (QueryPerformanceFrequency(&li) == 0)
	{
		mqsixLog("No performance freq counter!\n");
	}
	else
	{
		mqsixLog("hi/low of freq counter: %d:%d\n", li.HighPart, li.LowPart);
	}
*/
	mqsixAddAttribute(&pMyContext->attributes, L"measureType", L"none");
	mqsixAddAttribute(&pMyContext->attributes, L"timeStartLabel", L"none");
	mqsixBrokerInfo(nodeObject, &pMyContext->brokerInfo);
	return(pMyContext);
}

///////////////////////////////////////////////////////////
// DeleteNodeContext
static void DeleteNodeContext(CciContext *context)
{
	struct NodeContext	*pMyContext;

	pMyContext = (struct NodeContext *)context;
	mqsixLog(">> DeleteNodeContext\n");
	mqsixReleaseAttributes(pMyContext->attributes);
	pMyContext->attributes = NULL;
	if (pMyContext->label != NULL)
	{
		free(pMyContext->label);
		pMyContext->label = NULL;
	}
	free(context);
}

///////////////////////////////////////////////////////////
// GetAttributeName
static int GetAttributeName(
	CciContext	*context,
	int			index,
	CciChar		*buffer,
	int			bufsize)
{
	int	iRc;
	struct NodeContext	*pMyContext;

	pMyContext = (struct NodeContext *)context;
	iRc = mqsixGetAttributeName(
		pMyContext->attributes,
		index,
		buffer,
		bufsize);
	return(iRc);
}

///////////////////////////////////////////////////////////
// GetAttribute
static int GetAttribute(
	CciContext	*context,
	CciChar		*attrName,
	CciChar		*buffer,
	int bufsize)
{
	struct NodeContext	*pMyContext;
	int					iRc;

	pMyContext = (struct NodeContext *)context;
	iRc = mqsixGetAttribute(
		pMyContext->attributes,
		attrName,
		buffer,
		bufsize);
	return(iRc);
}

///////////////////////////////////////////////////////////
// SetAttribute
static int SetAttribute(
	CciContext	*context,
	CciChar		*attrName,
	CciChar		*attrValue)
{
	struct NodeContext	*pMyContext;
	int					iRc;

	pMyContext = (struct NodeContext *)context;
	if (mqsixStringCompare(attrName, L"label") == 0)
	{
		pMyContext->label = mqsixStringDuplicate(attrValue);
		mqsixLog("Set label -> %S\n", pMyContext->label);
		return(1);
	}
	iRc = mqsixSetAttribute(pMyContext->attributes,
		attrName,
		attrValue);

	if (mqsixStringCompare(attrName, L"measureType") == 0)
	{
		if (mqsixStringCompare(attrValue, L"ToEnd") == 0)
		{
			pMyContext->iMeasureType = TOEND;
		}
		if (mqsixStringCompare(attrValue, L"TimeStart") == 0)
		{
			pMyContext->iMeasureType = TIMESTART;
		}
		if (mqsixStringCompare(attrValue, L"TimeEnd") == 0)
		{
			pMyContext->iMeasureType = TIMEEND;
		}
	}
	return(iRc);
}

///////////////////////////////////////////////////////////
// GetEntryId
static char *GetEntryId(CciChar *strLabelU)
{
	char	*strLabelA;
	char	*strReturn;

	strReturn = (char *)malloc(MQSIX_MAX_STRING);
	strLabelA = mqsixUnicodeToASCII(strLabelU);
	sprintf(strReturn, "%s:%d:%d", strLabelA, OS_GetPid(), OS_GetThread());
	free(strLabelA);
	mqsixLog("Entry: %s\n", strReturn);
	return(strReturn);
}

///////////////////////////////////////////////////////////
// FindEntry
static NODE_TIME_ENTRY *FindEntry(CciChar *strLabelU)
{
	char	*strEntryId;
	int		i;
	strEntryId = GetEntryId(strLabelU);
	for (i=0; i<MAX_TIME_ENTRIES; i++)
	{
		if (entryList[i].used !=0 && strcmp(strEntryId, entryList[i].name) == 0)
		{
			free(strEntryId);
			return(&entryList[i]);
		}
	}
	free(strEntryId);
	return(NULL);
}

///////////////////////////////////////////////////////////
// AddEntry
static NODE_TIME_ENTRY *AddEntry(CciChar *strLabelU, MY_TIME_T time)
{
	int	i;
	char *strEntryId;

	strEntryId = GetEntryId(strLabelU);
	for (i=0; i<MAX_TIME_ENTRIES; i++)
	{
		if (entryList[i].used == 0)
		{
			entryList[i].used = 1;
			entryList[i].time = time;
			strcpy(entryList[i].name, strEntryId);
			free(strEntryId);
			return(&entryList[i]);
		}
	}
	free(strEntryId);
	return(NULL);
}

///////////////////////////////////////////////////////////
// GetTime
#ifdef _WIN32
static MY_TIME_T GetTime()
{
	MY_TIME_T	count;
	QueryPerformanceCounter((LARGE_INTEGER *)&count);
	return(count);
}
#else
static MY_TIME_T GetTime()
{
	MY_TIME_T	tv;
	gettimeofday(&tv, 0);
	return(tv);
}
#endif

///////////////////////////////////////////////////////////
// DiffTime
#ifdef _WIN32
static double DiffTime(MY_TIME_T start, MY_TIME_T end)
{
	unsigned long diffLong;
	__int64	freq;
	double			msecs;
	diffLong = (unsigned long)(end-start);
	QueryPerformanceFrequency((LARGE_INTEGER *)&freq);
	msecs = (double)diffLong*1000/freq;
	return(msecs);
}
#else
static double DiffTime(MY_TIME_T start, MY_TIME_T end)
{
  unsigned long val;
  if (end.tv_usec < start.tv_usec)
  {
      val = 1000000;
      val += start.tv_usec - end.tv_usec;
  }
  else
  {
      val = end.tv_usec - start.tv_usec;
  }
  val += (end.tv_sec - start.tv_sec)*1000000;
  return((double)val/1000.0);
}
#endif

///////////////////////////////////////////////////////////
// Evaluate
static void Evaluate(
	CciContext	*context, 
	CciMessage	*inDestinationList,
	CciMessage	*inExceptionList,
	CciMessage	*inMessage)
{
	int					iRc;
	struct NodeContext	*pMyContext;
	MY_TIME_T			startTime;
	MY_TIME_T			endTime;
	char				strTempA[20];
	char				*pStrFlow;
	char				*pStrStartLabel;
	char				*pStrLabelA;
	CciChar				strStartLabelU[MQSIX_MAX_STRING];		// Start label
	NODE_TIME_ENTRY		*pEntry;



	mqsixLog(">> Evaluate\n");
	mqsixDebugBreak();
	pMyContext = (struct NodeContext *)context;

	startTime = GetTime();
	switch(pMyContext->iMeasureType)
	{
	case TOEND:
		break;

	// TIMESTART:
	// Cache the time that this node was started.  Key used is label:pid:thread
	case TIMESTART:
		pEntry = AddEntry(pMyContext->label, startTime);
		mqsixLog("TIMESTART: Entry added for label: %s\n", pEntry->name);
		break;

	// TIMEEND:
	// Search the entries for a previous entry which matches the label:pid:thread
	// This entry will contain the time at which the corresponding node captured
	// the timing information.
	case TIMEEND:
		mqsixGetAttribute(pMyContext->attributes,
			L"timeStartLabel",
			strStartLabelU, sizeof(strStartLabelU));
		pEntry = FindEntry(strStartLabelU);
	// If pEntry == NULL then we failed to find a matching previous node that
	// recorded the start time.
		if (pEntry == NULL)
		{
			mqsixLog("Failed to match entry for label: %S\n", strStartLabelU);
			pStrStartLabel = mqsixUnicodeToASCII(strStartLabelU);
			pStrLabelA = mqsixUnicodeToASCII(pMyContext->label);
			pStrFlow = mqsixUnicodeToASCII(pMyContext->brokerInfo.messageFlowName);
			cciLog(&iRc, CCI_LOG_ERROR, __FILE__, __LINE__, "Evaluate",
				L"MQSIV2_EXECUTETIME", MSG_NO_MATCHING_LABEL,
				"(%1) [%2:%3] No matching node label '%4'",
			pStrFlow, pStrLabelA, pStrStartLabel, (char *)0);
			free(pStrStartLabel);
		}
		else	// ... we have good information ... so log it
		{
			sprintf(strTempA, "%.3f", DiffTime(pEntry->time, startTime));
			pStrLabelA = mqsixUnicodeToASCII(pMyContext->label);
			pStrFlow = mqsixUnicodeToASCII(pMyContext->brokerInfo.messageFlowName);
			pStrStartLabel = mqsixUnicodeToASCII(strStartLabelU);
mqsixLog("pStrFlow: %s pStrLabelA: %s\n", pStrFlow, pStrLabelA);
			cciLog(&iRc, CCI_LOG_INFORMATION, __FILE__, __LINE__, "Evaluate",
				L"MQSIV2_EXECUTETIME", MSG_TIME_END,
				"(%1) [%2:%3] Time through message flow from node '%4': %5 milliseconds",
				pStrFlow, pStrLabelA, pStrStartLabel, strTempA, (char *)0);
			free(pStrFlow);
			free(pStrLabelA);
			free(pStrStartLabel);
			mqsixLog("TIMEEND: from %S -- Time: %.3f\n",
				strStartLabelU, DiffTime(pEntry->time, startTime));
		}
		break;
	}

	// Propagate the message through the rest of the message flow.
	cniPropagate(&iRc, pMyContext->outTerminal, 
		inDestinationList, inExceptionList, inMessage);
	mqsixLog("-- Propagated through out terminal...\n");
	switch(pMyContext->iMeasureType)
	{
	// TOEND: - This is the post-propagate TIMEEND node.  We must now log the
	// difference in time between the start of the propagate and the current time.
	case TOEND:
		endTime = GetTime();
		sprintf(strTempA, "%.3f", DiffTime(startTime, endTime));
		pStrLabelA = mqsixUnicodeToASCII(pMyContext->label);
		pStrFlow = mqsixUnicodeToASCII(pMyContext->brokerInfo.messageFlowName);
		cciLog(&iRc, CCI_LOG_INFORMATION, __FILE__, __LINE__, "Evaluate",
			L"MQSIV2_EXECUTETIME", MSG_DELTA_TIME,
			"(%1) [%2:%3] Time through message flow: %4 milliseconds",
			pStrFlow, pStrLabelA, strTempA, (char *)0);
		free(pStrFlow);
		free(pStrLabelA);
		mqsixLog("-- Time: %.3f\n", DiffTime(startTime, endTime));
		break;

	// TIMESTART: - This is the post-propagate TIMESTART node.  We can release the
	// entry in the entry list now since it will never be reference again in this
	// message flow.
	case TIMESTART:
		pEntry->used = 0;
		break;

	// TIMEEND: - This is the post-propagate TIMEEND node.  Nothing need be done
	// here as it was logged prior to the propagate.
	case TIMEEND:
		break;
	}
	
	if (iRc != 0)
	{
		mqsixLog("An exception was reported on return from propogate through 'out': %d\n", iRc);
		cciRethrowLastException(&iRc);
		return;
	}
	mqsixLog("<< Evaluate\n");
	return;
}

///////////////////////////////////////////////////////////
// bipGetMessageflowNodeFactory
// 
// Entry point into the ExecuteTime node.  Called by the broker when
// the node is initially loaded.
/*
extern "C"
{
	void __declspec(dllexport) *bipGetMessageflowNodeFactory();
};
*/

void *bipGetMessageflowNodeFactory()
{
	static CNI_VFT	virtualFunctionTable;
	CciChar			factoryName[] = L"ExecuteTimeFactory";
	int				iRc;
	void			*factoryObject;
#ifdef _WIN32
	mqsixRegisterMessages(g_module, "MQSIV2_EXECUTETIME");
#endif
	mqsixLog("\n");
	mqsixLog("************************\n");
	mqsixLog("*** Loading Node LIL ***\n");
	mqsixLog("************************\n");
	mqsixLog(">> Creating nodeFactory: %S, lil build: %s, CCI Version: %d.%d\n",
		factoryName, __TIMESTAMP__, CCI_VERSION >> 16, CCI_VERSION & 0xFFFF);
	memset(&virtualFunctionTable, 0, sizeof(virtualFunctionTable));
	virtualFunctionTable.iFpCreateNodeContext	= CreateNodeContext;
	virtualFunctionTable.iFpDeleteNodeContext	= DeleteNodeContext;
	virtualFunctionTable.iFpGetAttributeName	= GetAttributeName;
	virtualFunctionTable.iFpSetAttribute		= SetAttribute;
	virtualFunctionTable.iFpGetAttribute		= GetAttribute;
	virtualFunctionTable.iFpEvaluate			= Evaluate;

	factoryObject = cniCreateNodeFactory(&iRc, factoryName);
	cniDefineNodeClass(&iRc, factoryObject, L"ExecuteTimeNode", &virtualFunctionTable); 

	return(factoryObject);
}
